/**
 * @file		lib_interface.c
 * @date		01/09/2022
 * @author		JackCarterSmith
 * @copyright	GPL-v3.0
 * @brief		Interface functions to external RSP libraries.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#if defined(_WIN32)
	#include <windows.h>
	#include <libloaderapi.h>
#else
	#include <dlfcn.h>
#endif
#include "options.h"
#include "lib_interface.h"


/*
 * Global handler declaration
 */
static void* hRSPTextureLib = NULL;
__RSPTexture_getVersion RSPTexture_getVersion;
__RSPTexture_createHMT RSPTexture_createHMT;
__RSPTexture_processHMTFile RSPTexture_processHMTFile;
__RSPTexture_getMaterialFromID RSPTexture_getMaterialFromID;
__RSPTexture_getMaterialName RSPTexture_getMaterialName;
__RSPTexture_getMaterialOpacity RSPTexture_getMaterialOpacity;
__RSPTexture_getMaterialAmbient RSPTexture_getMaterialAmbient;
__RSPTexture_freeHMT RSPTexture_freeHMT;


/**
 * Link functions proxies against RSPTextureLib dll/so.
 * @param opt_ptr Table of parameters to record module loading status.
 */
void* linkTextureLib(T_PROG_OPTIONS* opt_ptr) {
	if (hRSPTextureLib == NULL) {
#if defined(_WIN32)
		hRSPTextureLib = LoadLibraryA(RSPTEXTURE_OSLIBFILE);
		if (hRSPTextureLib) {
			RSPTexture_getVersion = (__RSPTexture_getVersion)GetProcAddress(
					hRSPTextureLib, RSPTEXTURE_FUNCNAME_GET_VERSION);
			RSPTexture_createHMT = (__RSPTexture_createHMT)GetProcAddress(
					hRSPTextureLib, RSPTEXTURE_FUNCNAME_NEW_HMT);
			RSPTexture_processHMTFile = (__RSPTexture_processHMTFile)GetProcAddress(
					hRSPTextureLib, RSPTEXTURE_FUNCNAME_PROC_HMT);
			RSPTexture_getMaterialFromID = (__RSPTexture_getMaterialFromID)GetProcAddress(
					hRSPTextureLib, RSPTEXTURE_FUNCNAME_GET_MAT);
			RSPTexture_getMaterialName = (__RSPTexture_getMaterialName)GetProcAddress(
					hRSPTextureLib, RSPTEXTURE_FUNCNAME_GET_NAME);
			RSPTexture_getMaterialOpacity = (__RSPTexture_getMaterialOpacity)GetProcAddress(
					hRSPTextureLib, RSPTEXTURE_FUNCNAME_GET_OPACITY);
			RSPTexture_getMaterialAmbient = (__RSPTexture_getMaterialAmbient)GetProcAddress(
					hRSPTextureLib, RSPTEXTURE_FUNCNAME_GET_AMBIENT);
			RSPTexture_freeHMT = (__RSPTexture_freeHMT)GetProcAddress(
					hRSPTextureLib, RSPTEXTURE_FUNCNAME_FREE_HMT);

			if (RSPTexture_getVersion && RSPTexture_createHMT && RSPTexture_processHMTFile
					&& RSPTexture_getMaterialFromID && RSPTexture_freeHMT && RSPTexture_getMaterialName
					&& RSPTexture_getMaterialOpacity && RSPTexture_getMaterialAmbient) {
				opt_ptr->texture_module = 1;
				printf("[INFO] Loaded RSPTextureLib v%s\n", RSPTexture_getVersion());
			} else {
				printf("[INFO] RSPTextureLib linking failure. Maybe incorrect version or file corrupted?\n");
			}
		} else {
			printf("[INFO] RSPTextureLib not found.\n");
		}
#else
		hRSPTextureLib = dlopen(RSPTEXTURE_OSLIBFILE, RTLD_LAZY | RTLD_DEEPBIND);
		if (hRSPTextureLib) {
			RSPTexture_getVersion = dlsym(
					hRSPTextureLib, RSPTEXTURE_FUNCNAME_GET_VERSION);
			RSPTexture_createHMT = dlsym(
					hRSPTextureLib, RSPTEXTURE_FUNCNAME_NEW_HMT);
			RSPTexture_processHMTFile = dlsym(
					hRSPTextureLib, RSPTEXTURE_FUNCNAME_PROC_HMT);
			RSPTexture_getMaterialFromID = dlsym(
					hRSPTextureLib, RSPTEXTURE_FUNCNAME_GET_MAT);
			RSPTexture_getMaterialName = dlsym(
					hRSPTextureLib, RSPTEXTURE_FUNCNAME_GET_NAME);
			RSPTexture_getMaterialOpacity = dlsym(
					hRSPTextureLib, RSPTEXTURE_FUNCNAME_GET_OPACITY);
			RSPTexture_getMaterialAmbient = dlsym(
					hRSPTextureLib, RSPTEXTURE_FUNCNAME_GET_AMBIENT);
			RSPTexture_freeHMT = dlsym(
					hRSPTextureLib, RSPTEXTURE_FUNCNAME_FREE_HMT);

			if (RSPTexture_getVersion && RSPTexture_createHMT && RSPTexture_processHMTFile
					&& RSPTexture_getMaterialFromID && RSPTexture_freeHMT && RSPTexture_getMaterialName
					&& RSPTexture_getMaterialOpacity && RSPTexture_getMaterialAmbient) {
				opt_ptr->texture_module = 1;
				printf("[INFO] Loaded RSPTextureLib v%s\n", RSPTexture_getVersion());
			} else {
				printf("[INFO] RSPTextureLib linking failure. Maybe incorrect version or file corrupted?\n");
			}
		} else {
			printf("[INFO] RSPTextureLib not found.\n");
		}
#endif
	}

	return hRSPTextureLib;
}

/**
 * Release libraries from dynamic loader.
 */
void unlinkTextureLib(void) {
	if (hRSPTextureLib == NULL) return;

#if defined(_WIN32)
	//FreeLibraryAndExitThread(hRSPTextureLib, 0);
	FreeLibrary(hRSPTextureLib);
#else
	dlclose(hRSPTextureLib);
#endif
}
