/**
 * @file		model_export.c
 * @date		27/07/2022
 * @author		JackCarterSmith
 * @copyright	GPL-v3.0
 * @brief		Export datas to Waveform OBJ format.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "options.h"
#include <RSPModel_datatypes.h>
#include <RSPModel_errordefs.h>
#include "obj/obj.h"
#include "lib_interface.h"
#include "model_export.h"


unsigned char exportOBJModel(T_RSPMODEL_OBJECT* hob_objects, const char *out_path, void* hmt_struct, T_PROG_OPTIONS* p_opts) {
	char export_path[1024];
	char objExport_name[128];
	char mtlExport_name[128];
	char material_name[64];
	obj* objConstruct = NULL;
	void* material = NULL;
	unsigned int i,j,k,a;
	unsigned int poly_count = 0;
	int surfID = 0, materialID = 0, tmpVertex = 0, tmpIndex = 0;
	float vertexBuff[3] = {0}, textureBuff[2] = {0}, materialAmbient[4] = {0};
	int indicesBuff[3] = {0};

	if (hob_objects == NULL || out_path == NULL)
		return RSPLIB_ERROR_ARGS_NULL;

#ifdef _WIN32
	snprintf(export_path, 1024, "%s-out\\", out_path);
#else
	snprintf(export_path, 1024, "%s-out/", out_path);
#endif
	snprintf(objExport_name, 128, "%s.obj", hob_objects->name);
	snprintf(mtlExport_name, 128, "%s.mtl", hob_objects->name);

	objConstruct = obj_create(NULL);

	//TODO: should be deconstructed in sub-functions
	// For each objects sub-part, create it's faces and surfaces/materials datas
	for ( i = 0; i < hob_objects->object_part_count; i++) {

		// Build faces datas: vertices, indices table, materials and UV mapping
		for ( j = 0; j < hob_objects->object_parts[i].face_count; j++ ) {

			// Construct materials list - add new when first loop or mt_index change
			if (j == 0 || ( hob_objects->object_parts[i].faces[j].mt_index != hob_objects->object_parts[i].faces[j-1].mt_index )) {
				surfID = obj_add_surf(objConstruct);

				if (p_opts->texture_module) {
					materialID = obj_add_mtrl(objConstruct);

					// Set material properties for face group
					material = RSPTexture_getMaterialFromID(hmt_struct, hob_objects->object_parts[i].faces[j].mt_index);
					obj_set_mtrl_name(objConstruct, materialID, RSPTexture_getMaterialName(material));
					materialAmbient[0] = RSPTexture_getMaterialOpacity(material);
					materialAmbient[1] = materialAmbient[0];
					materialAmbient[2] = materialAmbient[0];
					materialAmbient[3] = materialAmbient[0];
					//obj_set_mtrl_c(objConstruct, materialID, OBJ_KA, materialAmbient);
					obj_set_mtrl_c(objConstruct, materialID, OBJ_KD, materialAmbient);
					//obj_set_mtrl_c(objConstruct, materialID, OBJ_KS, materialAmbient);

					// Add image file name if textured
					if (hob_objects->object_parts[i].faces[j].flags_bits.fHasTexture) {
						snprintf(material_name, 64, "%s.png", RSPTexture_getMaterialName(material));
						//obj_set_mtrl_map(objConstruct, materialID, OBJ_KA, material_name);
						obj_set_mtrl_map(objConstruct, materialID, OBJ_KD, material_name);
						//obj_set_mtrl_map(objConstruct, materialID, OBJ_KS, material_name);
					}

					// Link material to surface
					obj_set_surf(objConstruct, surfID, materialID);
				}
			}

			// Add vertex corresponding to indices
			for ( k = 0; k < 3; k++ ) {
				tmpVertex = obj_add_vert(objConstruct);
				a = hob_objects->object_parts[i].faces[j].indices[k];

				vertexBuff[0] = ((float)1/1024) * -hob_objects->object_parts[i].vertices[a].x;		// Invert X to fix mirror display
				vertexBuff[1] = ((float)1/1024) * -hob_objects->object_parts[i].vertices[a].y;		// Invert Y to render upside up
				vertexBuff[2] = ((float)1/1024) * hob_objects->object_parts[i].vertices[a].z;
				obj_set_vert_v(objConstruct, tmpVertex, vertexBuff);

				if (hob_objects->object_parts[i].faces[j].flags_bits.fHasTexture) {
					textureBuff[0] = ((float)1/4096) * hob_objects->object_parts[i].faces[j].tex_coords[k].u;
					textureBuff[1] = -((float)1/4096) * hob_objects->object_parts[i].faces[j].tex_coords[k].v + 1;
					obj_set_vert_t(objConstruct, tmpVertex, textureBuff);
				}
			}
			// Build indices table for this face
			tmpIndex = obj_add_poly(objConstruct, surfID);
			indicesBuff[0] = 3 * poly_count;
			indicesBuff[1] = 3 * poly_count + 1;
			indicesBuff[2] = 3 * poly_count + 2;
			poly_count++;
			obj_set_poly(objConstruct, surfID, tmpIndex, indicesBuff);

			// Process 2 triangles if face is Quad
			if (hob_objects->object_parts[i].faces[j].flags_bits.fIsQuad) {
				for ( k = 0; k < 3; k++ ) {
					tmpVertex = obj_add_vert(objConstruct);
					a = hob_objects->object_parts[i].faces[j].indices[(k + 2) & 0x3];			// TODO: can't be written properly?

					vertexBuff[0] = ((float)1/1024) * -hob_objects->object_parts[i].vertices[a].x;		// Invert X to fix mirror display
					vertexBuff[1] = ((float)1/1024) * -hob_objects->object_parts[i].vertices[a].y;		// Invert Y to render upside up
					vertexBuff[2] = ((float)1/1024) * hob_objects->object_parts[i].vertices[a].z;
					obj_set_vert_v(objConstruct, tmpVertex, vertexBuff);

					if (hob_objects->object_parts[i].faces[j].flags_bits.fHasTexture) {
						textureBuff[0] = ((float)1/4096) * hob_objects->object_parts[i].faces[j].tex_coords[(k + 2) & 0x3].u;
						textureBuff[1] = -((float)1/4096) * hob_objects->object_parts[i].faces[j].tex_coords[(k + 2) & 0x3].v + 1;
						obj_set_vert_t(objConstruct, tmpVertex, textureBuff);
					}
				}
				// Build indices table for this face
				tmpIndex = obj_add_poly(objConstruct, surfID);
				indicesBuff[0] = 3 * poly_count;
				indicesBuff[1] = 3 * poly_count + 1;
				indicesBuff[2] = 3 * poly_count + 2;
				poly_count++;
				obj_set_poly(objConstruct, surfID, tmpIndex, indicesBuff);
			}
		}
	}

	if (p_opts->export_mtl) {
		if (p_opts->output_dir)
			obj_write(objConstruct, objExport_name, mtlExport_name, export_path, 6);
		else
			obj_write(objConstruct, objExport_name, mtlExport_name, NULL, 6);
	} else {
		if (p_opts->output_dir)
			obj_write(objConstruct, objExport_name, NULL, export_path, 6);
		else
			obj_write(objConstruct, objExport_name, NULL, NULL, 6);
	}

	obj_delete(objConstruct);

	return RSPLIB_SUCCESS;
}
